#!/usr/bin/env python
# Copyright 2008 David D Lowe
# released under the GPLv3
"""Contains lots of custom exceptions used in this project.
This module has been named ep_exceptions in order not to conflict with the exceptions module"""

class FileNotFoundException(Exception):
    def __init__(self, filename, pigment=None):
        self.filename = filename
        self.pigment = pigment
    def __str__(self):
        return "File not found: \"%s\" pigment: \"%s\"" % \
        (self.filename, str(self.pigment))

class PreviewFileNotFoundException(FileNotFoundException):
    def __str__(self):
        return "Preview file not found: \"%s\" pigment: \"%s\"" % \
        (self.filename, str(self.pigment))

class AttachmentNotFoundException(FileNotFoundException):
    def __str__(self):
        return "Attachment not found: \"%s\" pigment: \"%s\"" % \
        (self.filename, str(self.pigment))

class PigmentNotFoundException(Exception):
    def __init__(self, pigment, pigmentType, directories=None):
        self.pigment = pigment
        self.pigmentType = pigmentType
        self.directories = directories
    def __str__(self):
        return "Pigment not found: codename: %s type: %s directories: %s" % \
        (self.pigment, self.pigmentType, self.directories)

class PigmentTypeNotFoundException(Exception):
    def __init__(self, pigmentType):
        self.pigmentType = pigmentType
    def __str__(self):
        return "Pigment type unrecognised: \"%s\"" % self.pigmentType

class TooManyAttachmentsException(Exception):
    def __init__(self, number): self.number = number
    def __str__(self): return "Too many attachments: %s" % self.number

class PigmentVersionNotAccepted(Exception):
    def __init__(self, refusedVersion, acceptedVersions):
        self.refusedVersion = refusedVersion
        self.acceptedVersions = acceptedVersions
    def __str__(self):
        out = "Pigment version %s is not recognised. Accepted pigment versions are:" % str(self.refusedVersion)
        for ver in self.acceptedVersions:
            out = out + " \"" + str(ver) + "\""
        return out

class NoFirstTextError(Exception):
    def __init__(self, tagName=None): self.tagName = tagName
    def __str__(self):
        if not self.tagName is None:
            return "no text for tag " + self.tagName
        return Exception.__str__(self)


class MissingFileInPigment(Exception):
    def __init__(self, pigm, filename): 
        self.filename = filename
        self.pigm = pigm
    def __str__(self):
        return "Missing file : \"%s\" for pigment %s" % (self.filename, str(self.pigm))

class UrlNotFoundException(Exception):
    def __init__(self, url):
        self.url = url
    def __str__(self):
        return "Cannot access url : %s" % self.url

class RequireThreadTypeException(Exception):
    def __str__(self):
        import threading
        return "The current thread has the wrong name, current name: %s" % threading.currentThread().getName()

class PigmentMoreOldFormat(Exception):
    def __init__(self, pigment, attachment):
        self.pigment = pigment
        self.attachment = attachment
    def __str__(self):
        return "Pigment files need to be moved to expected location"