#!/usr/bin/env python
# David D Lowe (c) 2008-2009
# licensed under GPLv3
"""loadrepo.py provides loading and displaying of the repository's information
in a EpidermisApp's window"""

from __future__ import division
import const
from const import DEBUG, MY_CACHE_HOME, MY_DATA_HOME, PIGMENT_TYPES
import pygtk
pygtk.require("2.0")
import gtk, gobject,os,sys,threading,subprocess, gconf, xml.parsers
import urlparse, urllib, urllib2
import managepigments
from glob import glob
import handy
import ep_exceptions
from gettext import gettext as _
import urllib2
import traceback

_acceptedEpidermisVersions = ("0.1", "0.4")

def load_repo(application, forceReload=False):
    """Loads epidermis repository, downloads and displays it
    application: the EpidermisApp object
    forceReload: whether to ignore application.repoLoaded or not, if forceReload 
        is False, load_repo will will only activate if app.repoLoaded is False"""
    handy.require_main_thread()
    app = application
    if forceReload:
        app.repoLoaded = False

    if not app.repoLoaded:
        wTree = gtk.glade.XML(app.wTreeFileName)
        fetchWin = wTree.get_widget("winOperations")
        fetchWin.set_destroy_with_parent(True)
        progressbar = wTree.get_widget("progressbarOperations")
        cancelButton = wTree.get_widget("butOperationsCancel")
        progressbar.set_fraction(0)
        
        def fetchWin_delete_callback(widget, event):
            cancel_callback(cancelButton)
            return True
        handleridDelete = fetchWin.connect("delete-event", fetchWin_delete_callback)
        
        app.controls.tvPigments = {}
        for pt in PIGMENT_TYPES:
            app.controls.tvPigments[pt] = app.wTree.get_widget("treeviewRepo" + managepigments.get_pigment_type(pt).directoryName.capitalize())
            if app.controls.tvPigments[pt] is None:
                print >> sys.stderr, "Cannot find app.controls.tvPigments[" + pt + "]"
            
        app.cancel = False
        def cancel_callback(widget, data=None):
            """cancel_callback for the cancel button on the download window"""
            handy.require_non_main_thread()
            app.cancel = True
            gobject.idle_add(lambda:fetchWin.hide())
            gobject.idle_add(lambda:app.switch_to_toggled(app.controls.tgInstalled))
            gobject.idle_add(lambda:app.controls.tgInstalled.set_active(True))
            gobject.idle_add(lambda:fetchWin.disconnect(handleridDelete))
            gobject.idle_add(lambda:cancelButton.disconnect(handleridCancel))
        handleridCancel = cancelButton.connect("clicked", cancel_callback)
        
        class CancelException(Exception):
            """exception that may be raised by the download_hook function"""
            def __str__(self): return "CancelException"
        
        def download_hook(blocksDownloaded, blockSize, size):
            """download_hook for the download thread that displays progress
            in the progressbar"""
            handy.require_non_main_thread()
            downloaded = blocksDownloaded * blockSize
            frac = blocksDownloaded * blockSize/ size
            if frac > 1:
                frac = 1
                downloaded = size
            if frac < 0: frac = 0
            gtk.gdk.threads_enter()
            progressbar.set_fraction(frac)
            if size >= 0:
                progressbar.set_text(handy.bytes_in_str(downloaded) + "/" + handy.bytes_in_str(size))
            else:
                progressbar.set_text(handy.bytes_in_str(downloaded) + "/??")
            gtk.gdk.threads_leave()
            if app.cancel:
                raise(CancelException())
        
        def load_cached_repo():
            """loads repo from cache"""
            
            # read the downloaded preview_summary.xml
            handy.require_non_main_thread()
            filename = os.path.join(MY_CACHE_HOME, "preview_summary.xml")
            if not os.path.exists(filename):
                gobject.idle_add(lambda:app.show_error(_("No cached repo information, try again"), terminalText=_("No cached preview_summary.xml"),parent=fetchWin))
                clear_repo_cache(app)
            else:
                # preview_summary.xml exists
                # try to read and display the information
                try:
                    from xml.dom import minidom
                    thexml = minidom.parse(filename)
                except xml.parsers.expat.ExpatError:
                    app.show_error(_("Invalid repo information"), terminalText=_("Invalid xml: %s") % filename, parent=fetchWin)
                    clear_repo_cache(app)
                else:
                    # valid XML in preview_summary.xml
                    app.pigmentsData = {}
                    app.pigmentsToInstall = {}
                    for pt in PIGMENT_TYPES[1:] + ["skin"]:
                        app.pigmentsToInstall[pt] = {}
                        app.pigmentsData[pt] = app.PigmentTypeData(pt, app.installed_toggled, app.controls.tvPigments[pt])
                        items = thexml.getElementsByTagName(managepigments.get_pigment_type(pt).directoryName)[0]
                        for item in items.getElementsByTagName("item"):
                            try:
                                newPigmentInstalled = False
                                newPigmentCodename = ""
                                newPigmentCodename = handy.get_first_text(item.getElementsByTagName("codename")[0])
                                newPigmentInstalled = app.is_pigment_installed(pt, newPigmentCodename)
                            except IndexError, ee:
                                app.show_error(_("Missing element in item %(id)s in %(filename)s") % {"id":item,
                                    "filename":filename})
                            except ep_exceptions.NoFirstTextError, ee:
                                app.show_error(_("Missing text in xml of item %(id)s in %(filename)s") % {"id":itemId,
                                    "filename":filename})
                            newPigment = None
                            if newPigmentInstalled:
                                for myPigm in app.installedPigments[pt]:
                                    if myPigm.codeName == newPigmentCodename:
                                        newPigment = myPigm
                                        break
                                if newPigment is None:
                                    raise(Exception("Cannot find installed pigment: %(newPigmentCodename)s type %(type)s" 
                                        % {'newPigmentCodename':newPigmentCodename, 'type':pt}))
                                app.pigmentsData[pt].append_to_listStore(newPigment, installed=True)
                            else: # pigment is not installed
                                newPigment = managepigments.create_pigment(pt)
                                itemId = item.getAttribute("id")
                                try:
                                    newPigment.set_human_name(handy.get_first_text(item.getElementsByTagName("name")[0]))
                                    newPigment.codeName = newPigmentCodename
                                    newPigment.set_description(handy.get_first_text(item.getElementsByTagName("description")[0]))
                                    newPigment.author = handy.get_first_text(item.getElementsByTagName("author")[0])
                                except IndexError, ee:
                                    app.show_error(_("Missing element in item %(id)s in %(filename)s") % {"id":itemId,
                                        "filename":filename})
                                except ep_exceptions.NoFirstTextError, ee:
                                    app.show_error(_("Missing text in xml of item %(id)s in %(filename)s") % {"id":itemId,
                                        "filename":filename})
                                if len(item.getElementsByTagName("preview")) > 0:
                                    try:
                                        newPigment.set_preview_file(os.path.join(MY_CACHE_HOME, handy.get_first_text(item.getElementsByTagName("preview")[0])))
                                    except ep_exceptions.NoFirstTextError:
                                        print >> sys.stderr, "no first text for preview in %s" % filename
                                if pt != "skin":
                                    app.pigmentsData[pt].append_to_listStore(newPigment, newPigmentInstalled)
                                else:
                                    for ptd in PIGMENT_TYPES[1:]:
                                        if len(item.getElementsByTagName(ptd + "Dependancy")) > 0:
                                            try:
                                                linkedCodename = handy.get_first_text(item.getElementsByTagName(ptd + "Dependancy")[0])
                                            except ep_exceptions.NoFirstTextError:
                                                print >> sys.stderr, "no first text for %sDependancy in %s in %s" % (ptd, newPigment.codeName, filename)
                                            else:
                                                foundPigment = None
                                                for pt2 in PIGMENT_TYPES:
                                                    for pigm in app.pigmentsData[pt2].pigments:
                                                        if pigm.type.codeName == ptd and pigm.codeName == linkedCodename:
                                                            foundPigment = pigm
                                                            break
                                                if foundPigment is None:
                                                    print >> sys.stderr, "looking for pigment " + linkedCodename + " of type " + ptd
                                                    raise(Exception("can't find pigment"))
                                                newPigment.set_linked_pigment(foundPigment)
                                    app.pigmentsData[pt].append_to_listStore(newPigment, newPigmentInstalled)

                    app.originalPigmentsListStore = {}
                    for pp in app.pigmentsData:
                        app.originalPigmentsListStore[app.pigmentsData[pp].get_pigmentType()] = app.pigmentsData[pp].blank_listStore()
                        ls = app.originalPigmentsListStore[app.pigmentsData[pp].get_pigmentType()]
                        coun = 0
                        while (coun < app.pigmentsData[pp].get_listStore().iter_n_children(None)):
                            app.pigmentsData[pp].append_to_listStore(app.pigmentsData[pp].get_pigment(coun), listStore=ls, \
                                installed=app.pigmentsData[pp].get_installed(coun))    
                            coun += 1
                    app.repoLoaded = True
                    gobject.idle_add(lambda:app.controls.tgFindMore.set_active(True))
                    
        def prepare_downloaded_repo(filename):
            """this function extracts the downloaded repo archive in the
            appropriate places in the cache and then runs load_cached_repo()"""
            sFilename = filename.split("/")[-1:][0]
            subprocess.call(["rm", os.path.join(MY_CACHE_HOME, "preview_summary.xml")])
            subprocess.call(["mkdir", "-p", os.path.join(MY_CACHE_HOME)])
            subprocess.call(["mv", filename, os.path.join(MY_CACHE_HOME, sFilename)])
            subprocess.call(["gunzip", sFilename], cwd=MY_CACHE_HOME)
            subprocess.call(["tar", "-xf", sFilename[0:-3]], cwd=MY_CACHE_HOME, stdout=os.tmpfile())
            subprocess.call(["rm", os.path.join(MY_CACHE_HOME, sFilename[0:-3])])
            if not os.path.exists(os.path.join(MY_CACHE_HOME, "preview_summary.xml")):
                app.show_error(_("Invalid archive downloaded, does not contain preview_summary.xml"))
                clear_repo_cache(app)
            else:
                # preview_summary.xml exists
                try:
                    from xml.dom import minidom
                    thexml = minidom.parse(os.path.join(MY_CACHE_HOME, "preview_summary.xml"))
                except xml.parsers.expat.ExpatError:
                    app.show_error(terminalText="Invalid xml: " + os.path.join(MY_CACHE_HOME, "preview_summary.xml"), \
                        text="Invalid repo information", parent=fetchWin)
                    clear_repo_cache(app)
                else:
                    load_cached_repo()
                    app.checkOnlineRepo = False

            
                
        class DownloadRepoThread(threading.Thread):
            """thread that downloads the file using urllib.urlretrieve"""
            
            def __init__(self, url, hook, app, fetchWin):
                """url: url of repo file to download
                hook: hook function for urllib.urlretrieve
                app: EpidermisApp"""
                self.url = url
                self.hook = hook
                self.parent = app
                self.fetchWin = fetchWin
                #self.previewSummaryTgz = previewSummaryTgz
                threading.Thread.__init__(self)
                self.setName("DownloadRepoThread")
                self.repoURL = self.parent.repoURL
                
            def run(self):
                if DEBUG:
                    self.run2()
                else:
                    try:
                        self.run2()
                    except Exception, ee:
                        # TODO: better error handling
                        traceback.print_exc()
                        self.display_download_critical_error(str(ee), _("Error loading repo"), \
                            _("An unexpected error occured: %(errortext)s") % {"errortext":str(ee)})
                        raise
            
            def run2(self):
                if self.parent.checkOnlineRepo:
                    self.check_online()
                else:
                    gtk.gdk.threads_enter()
                    progressbar.set_text(_("Loading repo"))
                    load_cached_repo()
                    gobject.idle_add(lambda:self.fetchWin.hide())
                    gtk.gdk.threads_leave()
                
            def check_online(self):
                handy.require_non_main_thread()
                gtk.gdk.threads_enter()
                progressbar.set_text(_("Checking connection"))
                gtk.gdk.threads_leave()
                
                # Connect to the repo xml file
                req = urllib2.Request(self.repoURL)
                try:
                    repoXml = urllib2.urlopen(req)
                except urllib2.HTTPError, ee:
                    print "repoUrl", self.parent.repoURL, str(ee)
                    self.display_download_critical_error(_("Connection cannot be made to %s") % self.repoURL, \
                        friendlySecondaryText=_("Please make sure that the repo's URL is correct"))
                    try:
                        repoXml.close()
                    except: pass
                    return
                except urllib2.URLError, ee:
                    print >> sys.stderr, str(ee)
                    self.display_download_critical_error(_("Connection cannot be made to %s") % self.repoURL, \
                        friendlySecondaryText=_("Please verify your Internet connection"))
                    try:
                        repoXml.close()
                    except: pass
                    return
                
                from xml.dom import minidom
                thexml = minidom.parse(repoXml)
                root = thexml.getElementsByTagName("ghnsdownload")[0]
                validVer = False
                for ver in root.getAttribute("epidermisVersion").split(";"):
                    if ver in _acceptedEpidermisVersions:
                        validVer = True
                        break
                if not validVer:
                    self.display_download_critical_error(_("The repo's version (%s) is not supported by this version of Epidermis. Epidermis may be out of date") % root.getAttribute("epidermisVersion"))
                    print >> sys.stderr, str(ep_exceptions.PigmentVersionNotAccepted(root.getAttribute("epidermisVersion"), _acceptedEpidermisVersions))
                    return
                previewSummaryTgzXml = None
                for item in thexml.getElementsByTagName("stuff"):
                    if item.getAttribute("category") == "epidermis/repo_preview_summary":
                        previewSummaryTgzXml = handy.get_first_text(item.getElementsByTagName("payload")[0])
                if previewSummaryTgzXml is None:
                    print "self.parent.repoURL", self.parent.repoURL
                    raise(Exception("cannot find stuff element in XML for repo_preview_summary"))
                repoXml.close()
                server = urlparse.urlparse(self.parent.repoURL) # eg: server[1] = localhost
                if urlparse.urlparse(previewSummaryTgzXml)[1] == "":
                    serverPath = urlparse.urlparse(self.parent.repoURL)[2][: \
                        (-1*len(self.parent.repoURL.split("/")[-1]))] # eg: serverPath = /ghns
                    previewSummaryTgz = urlparse.urljoin(server[0] + "://", server[1])
                    previewSummaryTgz = urlparse.urljoin(previewSummaryTgz.replace("///", "//"), serverPath)
                    previewSummaryTgz = urlparse.urljoin(previewSummaryTgz, previewSummaryTgzXml)
                else: # absolute URL
                    previewSummaryTgz = previewSummaryTgzXml
                headers = {}
                if self.parent.repoLastModified != "":
                    headers["If-Modified-Since"] = self.parent.repoLastModified
                if self.parent.repoETag != "":
                    headers["If-None-Match"] = self.parent.repoETag
                req = urllib2.Request(previewSummaryTgz, headers=headers)
                usedCachedPreviews = False
                try:
                    previewSummaryFile = urllib2.urlopen(req)
                except urllib2.HTTPError, ee:
                    if ee.code == 304:
                        usedCachedPreviews = True
                    else:
                        print "repoUrl", self.parent.repoURL, str(ee)
                        self.display_download_critical_error(_("Connection cannot be made to %s") % previewSummaryTgz)
                        repoXml.close()
                        return
                except urllib2.URLError, ee:
                    print >> sys.stderr, str(ee)
                    raise
                    self.display_download_critical_error(_("Connection cannot be made to %s") % previewSummaryTgz, \
                        friendlySecondaryText=_("Please verify your Internet connection"))
                    repoXml.close()
                    return
                else:
                    firstSize = int(previewSummaryFile.info().getheader("Content-Length", "-1"))
                if usedCachedPreviews or previewSummaryFile.info().status == 304:
                    gtk.gdk.threads_enter()
                    load_cached_repo()
                    gtk.gdk.threads_leave()
                    gobject.idle_add(lambda:self.fetchWin.hide())
                else:
                    gtk.gdk.threads_enter()
                    progressbar.set_text(_("Downloading"))
                    gtk.gdk.threads_leave()
                    previewSummaryFile.close()
                    try:
                        filename, headers = urllib.urlretrieve(previewSummaryTgz, reporthook=self.hook)
                    except CancelException, e:
                        gobject.idle_add(lambda:self.parent.switch_to_toggled(self.parent.controls.tgInstalled))
                    except IOError, err:
                        self.display_download_critical_error(_("Error connecting to %(ps)s\n %(err)s") % {'ps':previewSummaryTgz,'err':err})
                    except socket.timeout, err:
                        self.display_download_critical_error(_("Error connecting to %(ps)s, timed out\n %(err)s") % {'ps':previewSummaryTgz, 'err':err})
                    else:
                        client = gconf.client_get_default()
                        if headers.has_key("Last-Modified"):
                            client.set_string("/apps/epidermis/repo_last_modified", headers["Last-Modified"])
                        else:
                            client.set_string("/apps/epidermis/repo_last_modified", "")
                        if headers.has_key("ETag"):
                            client.set_string("/apps/epidermis/repo_etag", headers["ETag"])
                        else:
                            client.set_string("/apps/epidermis/repo_etag", "")
                        gtk.gdk.threads_enter()
                        prepare_downloaded_repo(filename)
                        gtk.gdk.threads_leave()
                        gobject.idle_add(lambda:self.fetchWin.hide())
            
            # still under DownloadRepoThread
            def display_download_critical_error(self, errorText, friendlyErrorText=None, friendlySecondaryText=""):
                # TODO: make more friendly error displayer
                handy.require_non_main_thread()
                if friendlyErrorText is None:
                    friendlyErrorText = errorText
                print >> sys.stderr, errorText
                def dialog_response(dialog, response_id):
                    dialog.destroy()
                    gobject.idle_add(lambda:self.parent.switch_to_toggled(self.parent.controls.tgInstalled))
                gobject.idle_add(lambda:self.fetchWin.hide())
                gtk.gdk.threads_enter()
                downloadDialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, friendlyErrorText)
                downloadDialog.format_secondary_text(friendlySecondaryText)
                downloadDialog.connect("response", dialog_response)
                downloadDialog.set_skip_taskbar_hint(False)
                downloadDialog.set_title(_("Error"))
                downloadDialog.show()
                gtk.gdk.threads_leave()
                    
                    
        
        # we're still under if not app.repoLoaded
        fetchWin.show_all()

        app.currentDownloadRepoThread = DownloadRepoThread(app.repoURL, download_hook, app, fetchWin)
        app.currentDownloadRepoThread.start()

def clear_repo_cache(app):
    """deletes all cached information about the repository"""
    if not DEBUG:
        subprocess.call(["rm", "-r"] + glob(os.path.join(MY_CACHE_HOME, "*")))
    client = gconf.client_get_default()
    app.repoLastModified = ""
    client.set_string("/apps/epidermis/repo_last_modified","")
    app.repoETag = ""
    client.set_string("/apps/epidermis/repo_etag","")
    pass


def main():
    print "This file should not be run"
    sys.exit(1)

if __name__ == "__main__":
    main()
